/*----------------------------------------------------------------
	usbi2c.c - USB-I2C bridge console by Bonchan
		http://bonchan.justblog.jp/
	
	based on i2c_usb.c by Till Harbaum, thanks!
----------------------------------------------------------------*/

/*
 * i2c_usb.c - test application for the i2c-tiby-usb interface
 *             http://www.harbaum.org/till/i2c_tiny_usb
 *
 * $Id: i2c_usb.c,v 1.5 2007/01/05 19:30:43 harbaum Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
//#include <usb.h>
#include "usb.h"


#define LOOPS 100

#define I2C_TIMEOUT		1000

#define USB_CTRL_IN    (USB_TYPE_CLASS | USB_ENDPOINT_IN)
#define USB_CTRL_OUT   (USB_TYPE_CLASS)

/* the vendor and product id was donated by ftdi ... many thanks!*/
#define I2C_TINY_USB_VID  0x0403
#define I2C_TINY_USB_PID  0xc631

#ifdef WIN
#include <windows.h>
#include <winbase.h>
#define usleep(t) Sleep((t) / 1000)
#endif

#define I2C_M_RD		0x01

/* commands via USB, must e.g. match command ids firmware */
#define CMD_ECHO       0
#define CMD_GET_FUNC   1
#define CMD_SET_DELAY  2
#define CMD_GET_STATUS 3
#define CMD_I2C_IO     4
#define CMD_I2C_BEGIN  1  // flag to I2C_IO
#define CMD_I2C_END    2  // flag to I2C_IO

#define STATUS_IDLE          0
#define STATUS_ADDRESS_ACK   1
#define STATUS_ADDRESS_NAK   2

usb_dev_handle      *handle = NULL;


/* read a set of bytes from the i2c_tiny_usb device */
int i2c_tiny_usb_read(unsigned char cmd, void *data, int len) {
  int                 nBytes;

  /* send control request and accept return value */
  nBytes = usb_control_msg(handle, 
	   USB_CTRL_IN, 
	   cmd, 0, 0, data, len, 1000);

  if(nBytes < 0) {
    fprintf(stderr, "USB error: %s\n", usb_strerror());
    return nBytes;
  }

  return 0;
}

/* get i2c usb interface firmware version */
void i2c_tiny_usb_get_func(void) {
  unsigned long func;
  
  if(i2c_tiny_usb_read(CMD_GET_FUNC, &func, sizeof(func)) == 0)
    printf("Functionality = %lx\n", func);
}

/* set a value in the I2C_USB interface */
void i2c_tiny_usb_set(unsigned char cmd, int value) {
  if(usb_control_msg(handle, 
	     USB_TYPE_VENDOR, cmd, value, 0, 
	     NULL, 0, 1000) < 0) {
    fprintf(stderr, "USB error: %s\n", usb_strerror());
  }
}

/* get the current transaction status from the i2c_tiny_usb interface */
int i2c_tiny_usb_get_status(void) {
  int i;
  unsigned char status;
  
  if((i=i2c_tiny_usb_read(CMD_GET_STATUS, &status, sizeof(status))) < 0) {
    fprintf(stderr, "Error reading status\n");
    return i;
  }

  return status;
}


//----------------------------------------------------------------
// main
//
int main(int argc, char *argv[])
{
	struct usb_bus *bus;
	struct usb_device *dev;
	
	printf("USB-I2C bridge console (c) 2009 Bonchan\n");
	printf("-- http://bonchan.justblog.jp/\n");
	printf("This application based on i2c-tiby-usb interface by Till Harbaum\n");
	printf("-- http://www.harbaum.org/till/i2c_tiny_usb\n");
	printf("\n");
	
	usb_init();
	usb_find_busses();
	usb_find_devices();
	
	for(bus = usb_get_busses(); bus; bus = bus->next)
	{
		for(dev = bus->devices; dev; dev = dev->next)
		{
			if((dev->descriptor.idVendor == I2C_TINY_USB_VID) &&
				(dev->descriptor.idProduct == I2C_TINY_USB_PID))
			{
				printf("Found i2c_tiny_usb device on bus %s device %s.\n",
						bus->dirname, dev->filename);
				
				/* open device */
				if(!(handle = usb_open(dev)))
					fprintf(stderr, "Error: Cannot open the device: %s\n",
						usb_strerror());
				break;
			}
		}
	}
	
	if(!handle)
	{
		fprintf(stderr, "Error: Could not find i2c_tiny_usb device\n");
		printf("Press return to quit\n");
		getchar();
		return -1;
	}
	
	/* do some testing */
	i2c_tiny_usb_get_func();
	
	/* try to set i2c clock to 100kHz (10us), will actually result in ~50kHz */
	/* since the software generated i2c clock isn't too exact. in fact setting */
	/* it to 10us doesn't do anything at all since this already is the default */
	i2c_tiny_usb_set(CMD_SET_DELAY, 10);
	
	/*
		write
		> xx:yy yy yy
		
		read/write
		> xx:yy,4
	*/
	
	unsigned char sla = 1;
	
	while(1)
	{
		char cmd[128], *p, *q, *r;
		unsigned char buf[64];
		int a, wb, rb, i;
		
		printf(">");
		fgets(cmd, sizeof(cmd), stdin);
		cmd[sizeof(cmd) - 1] = 0;
		p = strchr(cmd, '\n');
		if(p) *p = 0;
		p = cmd;
		while(*p == ' ') p++;
		if(!*p) continue;
		
		if(tolower(*p) == 'q') break;
		
		a = strtol(p, &q, 16);
		while(*q == ' ') q++;
		if(*q != ':')
		{
serr:		fprintf(stderr, "Syntax Error\n");
			continue;
		}
		if(p != q)
		{
			if(a < 0 || a > 127) goto serr;
			sla = a;
		}
		p = q + 1;
		
		r = strchr(q, ',');
		if(r)
		{
			*(r++) = 0;
			while(*r == ' ') r++;
			rb = strtol(r, &r, 10);
			while(*r == ' ') r++;
			if(rb < 0 || *r) goto serr;
		}
		else
			rb = 0;
		
		// sla check
		if(sla == 0x00 && rb > 0)
		{
			fprintf(stderr, "General Call Address (0x00) can be used write mode only!\n");
			continue;
		}
		else if(sla <= 0x07 || sla >= 0x78)
		{
			fprintf(stderr, "Slave Address (0x%02x) is reserved!\n", sla);
			continue;
		}
		
		while(*p == ' ') p++;
		
		// fill buffer
		for(wb = 0; *p; wb++)
		{
			buf[wb] = strtol(p, &q, 16);
			if(p == q) goto serr;
			p = q;
			while(*p == ' ') p++;
		}
		
		if(!wb && !rb) continue;
		
		// start condition & write bytes
		if(wb && usb_control_msg(handle, USB_CTRL_OUT,
				CMD_I2C_IO + CMD_I2C_BEGIN + (rb ? 0 : CMD_I2C_END),
				0, sla, buf, wb, I2C_TIMEOUT) < wb)
		{
cerr:		fprintf(stderr, "Communication Error!\n");
			continue;
		}
		
		if(wb && (i2c_tiny_usb_get_status() != STATUS_ADDRESS_ACK))
		{
			fprintf(stderr, "Write command failed!\n");
			continue;
		}
		
		// (re-)start condition & read bytes
		if(rb && usb_control_msg(handle, USB_CTRL_IN,
				CMD_I2C_IO + (wb ? 0 : CMD_I2C_BEGIN) + CMD_I2C_END,
				I2C_M_RD, sla, buf, rb, I2C_TIMEOUT) < rb)
			goto cerr;
		
		if(rb && (i2c_tiny_usb_get_status() != STATUS_ADDRESS_ACK))
		{
			fprintf(stderr, "Read command failed!\n");
			continue;
		}
		
		// print read data
		for(i = 0; i < rb; i++) printf("%02x ", buf[i]);
		if(rb) printf("\n");
	}
	
	usb_close(handle);
	
	return 0;
}

